@@ -19,9 +19,9 @@
 
 """Provides the default implementation for flat review for Orca."""
 
-__id__        = "$Id: flat_review.py 4364 2008-11-22 21:42:36Z joanied $"
-__version__   = "$Revision: 4364 $"
-__date__      = "$Date: 2008-11-22 16:42:36 -0500 (Sat, 22 Nov 2008) $"
+__id__        = "$Id$"
+__version__   = "$Revision$"
+__date__      = "$Date$"
 __copyright__ = "Copyright (c) 2005-2008 Sun Microsystems Inc."
 __license__   = "LGPL"
 
@@ -36,7 +36,7 @@ import rolenames
 import settings
 
 from orca_i18n import _         # for gettext support
-from orca_i18n import Q_        # to provide qualified translatable strings
+from orca_i18n import C_        # to provide qualified translatable strings
 
 # [[[WDW - HACK Regular expression to split strings on whitespace
 # boundaries, which is what we'll use for word dividers instead of
@@ -45,6 +45,8 @@ from orca_i18n import Q_        # to provide qualified translatable strings
 #
 whitespace_re = re.compile(r'(\s+)', re.DOTALL | re.IGNORECASE | re.M)
 
+EMBEDDED_OBJECT_CHARACTER = u'\ufffc'
+
 class Char:
     """Represents a single char of an Accessibility_Text object."""
 
@@ -376,16 +378,12 @@ class StateZone(Zone):
                     # Translators: this is in reference to a radio button being
                     # selected or not.
                     #
-                    # ONLY TRANSLATE THE PART AFTER THE PIPE CHARACTER |
-                    #
-                    speechState = Q_("radiobutton|selected")
+                    speechState = C_("radiobutton", "selected")
                 else:
                     # Translators: this is in reference to a radio button being
                     # selected or not.
                     #
-                    # ONLY TRANSLATE THE PART AFTER THE PIPE CHARACTER |
-                    #
-                    speechState = Q_("radiobutton|not selected")
+                    speechState = C_("radiobutton", "not selected")
                 brailleState = \
                     settings.brailleRadioButtonIndicators[stateCount]
 
@@ -780,8 +778,7 @@ class Context:
         #print "LOOKING AT '%s'" % unicodeString
         for i in range(0, len(unicodeString) + 1):
             if (i != len(unicodeString)) \
-               and (unicodeString[i] != \
-                        orca_state.activeScript.EMBEDDED_OBJECT_CHARACTER):
+               and (unicodeString[i] != EMBEDDED_OBJECT_CHARACTER):
                 substringEndOffset += 1
             elif (substringEndOffset == substringStartOffset):
                 substringStartOffset += 1
@@ -859,10 +856,45 @@ class Context:
         else:
             zones = []
 
-        length = text.characterCount
-
         offset = 0
         lastEndOffset = -1
+        upperMax = lowerMax = text.characterCount
+        upperMid = lowerMid = upperMax /2
+        upperMin = lowerMin = 0
+        upperY = lowerY = 0
+        oldMid = 0
+
+        # performing binary search to locate first line inside clipped area
+        while oldMid != upperMid:
+            oldMid = upperMid
+            [x, y, width, height] = text.getRangeExtents(upperMid,
+                                                         upperMid+1,
+                                                         0)
+            upperY = y
+            if y > cliprect.y:
+                upperMax = upperMid
+            else:
+                upperMin = upperMid
+            upperMid = (upperMax - upperMin) /2 + upperMin
+
+        # performing binary search to locate last line inside clipped area
+        oldMid = 0
+        limit = cliprect.y+cliprect.height
+        while oldMid != lowerMid:
+            oldMid = lowerMid
+            [x, y, width, height] = text.getRangeExtents(lowerMid,
+                                                         lowerMid+1,
+                                                         0)
+            lowerY = y
+            if y > limit:
+                lowerMax = lowerMid
+            else:
+                lowerMin = lowerMid
+            lowerMid = (lowerMax - lowerMin) /2 + lowerMin
+
+        # finding out the zones
+        offset = upperMin
+        length = lowerMax
         while offset < length:
 
             [string, startOffset, endOffset] = text.getTextAtOffset(
@@ -1445,6 +1477,24 @@ class Context:
         #print "Current line=%d zone=%d word=%d char=%d" \
         #      % (lineIndex, zoneIndex, wordIndex, charIndex)
 
+    def routeToCurrent(self):
+        """Routes the mouse pointer to the current accessible."""
+
+        if (not self.lines) \
+           or (not self.lines[self.lineIndex].zones):
+            return
+
+        [string, x, y, width, height] = self.getCurrent(Context.CHAR)
+        try:
+            # We try to move to the left of center.  This is to
+            # handle toolkits that will offset the caret position to
+            # the right if you click dead on center of a character.
+            #
+            x = max(x, x + (width / 2) - 1)
+            eventsynthesizer.routeToPoint(x, y + height / 2, "abs")
+        except:
+            debug.printException(debug.LEVEL_SEVERE)
+
     def clickCurrent(self, button=1):
         """Performs a mouse click on the current accessible."""
 
